import ctypes
import pickle
from pathlib import Path
from typing import Callable, Tuple

import numpy as np

from multivar_horner import c_evaluation
from multivar_horner.c_evaluation import COMPILED_C_ENDING, compile_c_file, get_compiler, write_c_file
from multivar_horner.classes.abstract_poly import AbstractPolynomial
from multivar_horner.classes.factorisation import (
    BasePolynomialNode,
    HeuristicFactorisationRoot,
    OptimalFactorisationRoot,
)
from multivar_horner.classes.helpers import FactorContainer
from multivar_horner.global_settings import BOOL_DTYPE, FLOAT_DTYPE, PATH2CACHE, TYPE_1D_FLOAT, UINT_DTYPE
from multivar_horner.helper_fcts import rectify_query_point, validate_query_point
from multivar_horner.helpers_fcts_numba import eval_recipe


class HornerMultivarPolynomial(AbstractPolynomial):
    """a representation of a multivariate polynomial using Horner factorisation

    after computing the factorised representation of the polynomial,
    the instructions required for evaluation are being compiled and stored.
    The default format is C instructions.
    When there is no C compiler installed, the fallback option is encoding the required instructions
    in numpy arrays (here referred to as "recipes"),
    which can be processed by (``numba``) just in time compiled functions.

    Args:
        coefficients: ndarray of floats with shape (N,1)
            representing the coefficients of the monomials
            NOTE: coefficients with value 0 and 1 are allowed and will not affect the internal representation,
            because coefficients must be replaceable
        exponents: ndarray of unsigned integers with shape (N,m)
            representing the exponents of the monomials
            where m is the number of dimensions (self.dim),
            the ordering corresponds to the ordering of the coefficients, every exponent row has to be unique!
        rectify_input: bool, default=False
            whether to convert coefficients and exponents into compatible numpy arrays
            with this set to True, coefficients and exponents can be given in standard python arrays
        compute_representation: bool, default=False
            whether to compute a string representation of the polynomial
        verbose: bool, default=False
            whether to print status statements

        keep_tree: whether the factorisation tree object should be kept in memory after finishing factorisation
        store_c_instr: whether a C file with all required evaluation instructions
            should be created under any circumstances.
            By default the class will use C based evaluation, but skip if no compiler (``gcc``/``cc``) is installed.
        store_numpy_recipe: whether a pickle file with all required evaluation instructions
            in the custom ``numpy+numba`` format should be created under any circumstances.
            By default the class will use C based evaluation and only use this evaluation format as fallback.

    Attributes:
        num_monomials: the amount of coefficients/monomials N of the polynomial
        dim: the dimensionality m of the polynomial
            NOTE: the polynomial needs not to actually depend on all m dimensions
        unused_variables: the dimensions the polynomial does not depend on
        num_ops: the amount of mathematical operations required to evaluate the polynomial in this representation
        representation: a human readable string visualising the polynomial representation

        total_degree: the usual notion of degree for a polynomial.
            = the maximum sum of exponents in any of its monomials
            = the maximum l_1-norm of the exponent vectors of all monomials
            in contrast to 1D polynomials, different concepts of degrees exist for polynomials in multiple dimensions.
            following the naming in [1] L. Trefethen, “Multivariate polynomial approximation in the hypercube”,
            Proceedings of the American Mathematical Society, vol. 145, no. 11, pp. 4837–4844, 2017.
        euclidean_degree: the maximum l_2-norm of the exponent vectors of all monomials.
            NOTE: this is not in general an integer
        maximal_degree: the largest exponent in any of its monomials
            = the maximum l_infinity-norm of the exponent vectors of all monomials

        factorisation: a tuple of factorisation_tree and factor_container. s. below
        factorisation_tree: the object oriented, recursive data structure representing the factorisation
            (only if keep_tree=True)
        factor_container: the object containing all (unique) factors of the factorisation (only if keep_tree=True)
        root_value_idx: the index in the value array where the value of this polynomial
            (= root of the factorisation_tree) will be stored
        value_array_length: the amount of addresses (storage) required to evaluate the polynomial.
            for evaluating the polynomial in tree form intermediary results have to be stored in a value array.
            the value array begins with the coefficients of the polynomial.
            (without further optimisation) every factor requires its own address.

        copy_recipe: ndarray encoding the operations required to evaluate all scalar factors with exponent 1
        scalar_recipe: ndarray encoding the operations required to evaluate all remaining scalar factors
        monomial_recipe: ndarray encoding the operations required to evaluate all monomial factors
        tree_recipe: ndarray encoding the addresses required to evaluate
            the polynomial values of the factorisation_tree.
        tree_ops: ndarray encoding the type of operation required to evaluate
            the polynomial values of the factorisation_tree.
            encoded as a boolean ndarray separate from tree_recipe,
            since only the two operations ADD & MUL need to be encoded.

    Raises:
        TypeError: if coefficients or exponents are not given as ndarrays
            of the required dtype
        ValueError: if coefficients or exponents do not have the required shape or
            do not fulfill the other requirements
    """

    # __slots__ declared in parents are available in child classes. However, child subclasses will get a __dict__
    # and __weakref__ unless they also define __slots__ (which should only contain names of any additional slots).
    __slots__ = [
        "factorisation",
        "root_value_idx",
        "value_array_length",
        "keep_tree",
        "use_c_eval",
        "recipe",
        "_c_eval_fct",
        "ctype_x",
        "ctype_coeff",
        "root_class",
    ]

    # FIXME: creates duplicate entries in Sphinx autodoc
    def __init__(
        self,
        coefficients,
        exponents,
        rectify_input: bool = False,
        compute_representation: bool = False,
        verbose: bool = False,
        keep_tree: bool = False,
        store_c_instr: bool = False,
        store_numpy_recipe: bool = False,
        *args,
        **kwargs,
    ):
        super(HornerMultivarPolynomial, self).__init__(
            coefficients, exponents, rectify_input, compute_representation, verbose
        )
        self.root_class = HeuristicFactorisationRoot
        self.keep_tree: bool = keep_tree
        self.value_array_length: int
        self.recipe: Tuple

        self._c_eval_fct: Callable
        double = c_evaluation.C_TYPE_DOUBLE
        self.ctype_x = double * self.dim
        self.ctype_coeff = double * self.num_monomials

        self._configure_evaluation(store_c_instr, store_numpy_recipe)
        self.compute_string_representation(*args, **kwargs)

        if not self.keep_tree:
            try:
                del self.factorisation
            except AttributeError:
                pass

    def _configure_evaluation(
        self,
        store_c_instr: bool = False,
        store_numpy_recipe: bool = False,
    ):
        self.use_c_eval: bool = True
        if store_c_instr:
            # force storing a C file
            self._compile_c_file()
        if store_numpy_recipe:
            # force storing the numpy+Numba recipe
            self._compile_recipes()

        if not (store_c_instr or store_numpy_recipe):
            # by default use faster C evaluation -> compile C file
            # if the compilers are not installed this will raise a `ValueError`
            try:
                self._compile_c_file()
            except ValueError as exc:
                self.print(exc)
                # if C compilation does not work, use recipe (numpy+Numba based) evaluation as a fallback
                self.use_c_eval = False
                self._compile_recipes()

    def _compute_factorisation(self):
        # do not compute the factorisation when it is already present
        try:
            self.factorisation
            self.print("factorisation already exists.")
            return
        except AttributeError:
            pass

        self.print("computing factorisation...")
        # NOTE: do NOT automatically create all scalar factors with exponent 1
        # (they might be unused, since the polynomial must not actually depend on all variables)
        factor_container = FactorContainer()
        factorisation_tree: BasePolynomialNode = self.root_class(self.exponents, factor_container)
        self.root_value_idx = factorisation_tree.value_idx
        self.value_array_length = self.num_monomials + len(factor_container)
        self.factorisation: Tuple[BasePolynomialNode, FactorContainer] = (factorisation_tree, factor_container)

    @property
    def factorisation_tree(self) -> BasePolynomialNode:
        return self.factorisation[0]

    @property
    def factor_container(self) -> FactorContainer:
        return self.factorisation[1]

    def compute_string_representation(
        self,
        coeff_fmt_str: str = "{:.2}",
        factor_fmt_str: str = "x_{dim}^{exp}",
        *args,
        **kwargs,
    ) -> str:
        repre = ""
        if not self.compute_representation:
            self.representation = repre
            return repre
        repre = f"[#ops={self.num_ops}] p(x)"
        try:
            tree = self.factorisation_tree
            repre += " = " + tree.get_string_representation(self.coefficients, coeff_fmt_str, factor_fmt_str)
            # exponentiation with 1 won't cause an operation in this representation
            # but are present in the string representation due to string formatting restrictions
            # -> they should not be displayed (misleading)
            repre = repre.replace("^1", "")  # <- workaround for the default string format
        except AttributeError:
            pass  # self.factorisation_tree does not exist
        self.representation = repre
        return self.representation

    def _compile_recipes(self):
        """encode all instructions needed for evaluating the polynomial in 'recipes'

        recipes are represented as numpy ndarrays (cf. assembler instructions)

        -> acquire a data structure representing the factorisation tree
        -> avoid recursion and function call overhead during evaluation
        -> enables the use of jit compiled functions

        the factor container must now contain all unique factors used in the chosen factorisation
        during evaluation of a polynomial the values of all the factors are needed at least once
        -> compute the values of all factors once and store them
        -> store a pointer to the computed value for every factor ('value index' = address in the value array)
        this is required for compiling evaluation instructions depending on the factor values
        monomial factors exist only if their value is required during the evaluation of the parent polynomial
        scalar factors exist only if their value is required during the evaluation of existing monomial factors
        (scalar factors can be 'standalone' factors as well)
        -> values must not be overwritten (reusing addresses), because they might be needed again by another factor
        -> (without further optimisation) each factor requires its own space in the value array

        """
        # reuse pickled recipe if present
        try:
            self._load_recipe()
            return
        except ValueError:
            self.print("recipe does not exist yet.")
            pass

        # for compiling the recipes the factorisation must be computed first
        self._compute_factorisation()
        # the values of the factors are being stored after the coefficients
        # start the address assignment with the correct offset
        value_idx = self.num_monomials

        # compile the recipes for computing the value of all factors
        copy_recipe = []  # skip computing factors with exp 1, just copy x value
        scalar_recipe = []
        monomial_recipe = []
        self.num_ops = 0

        # count the amount of multiplications encoded by the recipes
        # NOTE: count exponentiations as exponent-1 multiplications, e.g. x^3 <-> 2 operations

        # -> IMPORTANT: value idx assignment must happen first for the scalar factors
        for scalar_factor in self.factor_container.scalar_factors:
            scalar_factor.value_idx = value_idx
            value_idx += 1
            copy_instr, scalar_instr = scalar_factor.get_recipe()
            copy_recipe += copy_instr
            scalar_recipe += scalar_instr
            if len(scalar_instr) > 0:
                exponent = scalar_instr[0][2]
                self.num_ops += exponent - 1

        for monomial_factor in self.factor_container.monomial_factors:
            monomial_factor.value_idx = value_idx
            value_idx += 1
            monomial_factor.factorisation_idxs = [f.value_idx for f in monomial_factor.scalar_factors]
            monomial_instr = monomial_factor.get_recipe()
            monomial_recipe += monomial_instr
            self.num_ops += 1  # every monomial instruction encodes one multiplication

        # compile the recipe for evaluating the Horner factorisation tree
        tree_recipe, tree_ops = self.factorisation_tree.get_recipe()
        # convert the recipes into the data types expected by the jit compiled functions
        # and store them
        tree_ops = np.array(tree_ops, dtype=BOOL_DTYPE)
        self.num_ops += len(tree_ops) - np.count_nonzero(tree_ops)  # every 0/False encodes a multiplication

        copy_recipe = np.array(copy_recipe, dtype=UINT_DTYPE).reshape((-1, 2))
        scalar_recipe = np.array(scalar_recipe, dtype=UINT_DTYPE).reshape((-1, 3))
        monomial_recipe = np.array(monomial_recipe, dtype=UINT_DTYPE).reshape((-1, 3))
        tree_recipe = np.array(tree_recipe, dtype=UINT_DTYPE).reshape((-1, 2))
        # IMPORTANT: strict ordering required!
        self.recipe = (copy_recipe, scalar_recipe, monomial_recipe, tree_recipe, tree_ops)
        self._pickle_recipe()

    def _pickle_recipe(self):
        path = self.recipe_file
        self.print(f'storing recipe in file "{path}"')
        pickle_obj = (self.recipe, self.num_ops, self.value_array_length, self.root_value_idx)
        with open(path, "wb") as f:
            pickle.dump(pickle_obj, f)

    def _load_recipe(self):
        path = self.recipe_file
        if not path.exists():
            raise ValueError("recipe pickle file does not exist.")
        self.print(f'loading recipe from file "{path}"')
        with open(path, "rb") as f:
            pickle_obj = pickle.load(f)
        self.recipe, self.num_ops, self.value_array_length, self.root_value_idx = pickle_obj

    def eval(self, x: TYPE_1D_FLOAT, rectify_input: bool = False) -> float:
        """computes the value of the polynomial at query point x

        either uses C  or numpy+Numba evaluation

        Args:
            x: ndarray of floats with shape = [self.dim] representing the query point
            rectify_input: whether to convert coefficients and exponents into compatible numpy arrays
                with this set to True, the query point x can be given in standard python arrays

        Returns:
             the value of the polynomial at point x

        Raises:
            TypeError: if x is not given as ndarray of dtype float
            ValueError: if x does not have the shape ``[self.dim]``
        """
        if rectify_input:
            x = rectify_query_point(x)

        # use numpy+Numba recipe evaluation as fallback
        # input type and shape should always be validated.
        #   otherwise the numba jit compiled functions may fail with cryptic error messages
        validate_query_point(x, self.dim)

        if self.use_c_eval:
            return self._eval_c(x)

        return self._eval_recipe(x)

    def _eval_recipe(self, x: TYPE_1D_FLOAT) -> float:
        """computes the value of the polynomial at query point x

        makes use of fast ``Numba`` just in time compiled functions
        """
        value_array = np.empty(self.value_array_length, dtype=FLOAT_DTYPE)
        # the coefficients are being stored at the beginning of the value array
        # TODO remove flatten, always store coefficients as a 1D array (also for horner fact.)?!
        #   also in MultivarPolynomial.eval()
        value_array[: self.num_monomials] = self.coefficients.flatten()
        return eval_recipe(
            x,
            value_array,
            *self.recipe,
            self.root_value_idx,
        )

    @property
    def c_eval_fct(self):
        try:
            return self._c_eval_fct
        except AttributeError:
            compiled_file = self.c_file_compiled
            cdll = ctypes.CDLL(str(compiled_file))
            fct = getattr(cdll, c_evaluation.EVAL_FCT)
            fct.argtypes = (self.ctype_x, self.ctype_coeff)
            fct.restype = c_evaluation.C_TYPE_DOUBLE
            self._c_eval_fct = fct

        return self._c_eval_fct

    def _eval_c(self, x: TYPE_1D_FLOAT) -> float:
        compiled_file = self.c_file_compiled
        if not compiled_file.exists():
            raise ValueError(f"missing compiled C file: {compiled_file}")

        x_typed = self.ctype_x(*x)
        coeffs = self.coefficients.flatten()
        coeffs_typed = self.ctype_coeff(*coeffs)

        p_x = self.c_eval_fct(x_typed, coeffs_typed)
        return p_x

    def get_c_file_name(self, ending: str = ".c") -> str:
        return f"eval_poly_{hash(self)}{ending}"

    @property
    def c_file(self) -> Path:
        return PATH2CACHE / self.get_c_file_name()

    @property
    def c_file_compiled(self):
        return PATH2CACHE / self.get_c_file_name(ending=COMPILED_C_ENDING)

    @property
    def recipe_file(self) -> Path:
        return PATH2CACHE / f"numpy_recipe_{hash(self)}.pickle"

    def _write_c_file(self):
        path_out = self.c_file
        if path_out.exists():
            return
        # for compiling the instructions the factorisation must be computed first
        self._compute_factorisation()
        self.print("compiling C instructions ...")
        try:
            factorisation = self.factorisation
        except AttributeError:
            raise ValueError(
                "need a stored factorisation tree, but the reference to it has already been deleted. "
                "initialise class with 'keep_tree=True`"
            )
        self.num_ops = write_c_file(self.num_monomials, self.dim, path_out, factorisation, self.verbose)

    def get_c_instructions(self) -> str:
        path = self.c_file
        self.print(f"reading C instructions from file {path}")
        with open(path) as c_file:
            instr = c_file.read()
        return instr

    def _compile_c_file(self):
        path_out = self.c_file_compiled
        if path_out.exists():
            self.print(f"using existing compiled C file {path_out}")
            return

        # NOTE: skip compiling C instructions if no compiler is available
        compiler = get_compiler()
        self._write_c_file()
        path_in = self.c_file
        self.print(f"compiling to file {path_out}")
        compile_c_file(compiler, path_in, path_out)


class HornerMultivarPolynomialOpt(HornerMultivarPolynomial):
    """a Horner factorised polynomial with an optimal factorisation found by searching all possible factorisations

    Optimality in this context refers to the minimal amount of operations needed for evaluation
    in comparison to other Horner factorisation (not other factorisatio/optimisation techniques).

    NOTES:

    * this requires MUCH more computational resources than just trying one factorisation
        (the number of possible factorisations is growing exponentially with the size of the polynomial!).

    * for the small polynomial examples in the current tests, the found factorisations were not superior
    """

    root_class = OptimalFactorisationRoot



from multivar_horner.classes.horner_poly import AbstractPolynomial
from multivar_horner.global_settings import TYPE_1D_FLOAT, TYPE_2D_INT
from multivar_horner.helper_fcts import rectify_query_point, validate_query_point
from multivar_horner.helpers_fcts_numba import count_num_ops_naive, naive_eval


class MultivarPolynomial(AbstractPolynomial):
    """a representation of a multivariate polynomial in 'canonical form' (without any factorisation)

    Args:
        coefficients: ndarray of floats with shape (N,1)
            representing the coefficients of the monomials
            NOTE: coefficients with value 0 and 1 are allowed and will not affect the internal representation,
            because coefficients must be replaceable
        exponents: ndarray of unsigned integers with shape (N,m)
            representing the exponents of the monomials
            where m is the number of dimensions (self.dim),
            the ordering corresponds to the ordering of the coefficients, every exponent row has to be unique!
        rectify_input: bool, default=False
            whether to convert coefficients and exponents into compatible numpy arrays
            with this set to True, coefficients and exponents can be given in standard python arrays
        compute_representation: bool, default=False
            whether to compute a string representation of the polynomial
        verbose: bool, default=False
            whether to print status statements

    Attributes:
        num_monomials: the amount of coefficients/monomials N of the polynomial
        dim: the dimensionality m of the polynomial
            NOTE: the polynomial needs not to actually depend on all m dimensions
        unused_variables: the dimensions the polynomial does not depend on
        num_ops: the amount of mathematical operations required to evaluate the polynomial in this representation
        representation: a human readable string visualising the polynomial representation

    Raises:
        TypeError: if coefficients or exponents are not given as ndarrays
            of the required dtype
        ValueError: if coefficients or exponents do not have the required shape or
            do not fulfill the other requirements or ``rectify_input=True`` and there are negative exponents
    """

    def __init__(
        self,
        coefficients: TYPE_1D_FLOAT,
        exponents: TYPE_2D_INT,
        rectify_input: bool = False,
        compute_representation: bool = False,
        verbose: bool = False,
        *args,
        **kwargs,
    ):

        super(MultivarPolynomial, self).__init__(
            coefficients, exponents, rectify_input, compute_representation, verbose
        )

        # NOTE: count the number of multiplications of the representation
        # not the actual amount of operations required by the naive evaluation with numpy arrays
        self.num_ops = count_num_ops_naive(self.exponents)
        self.compute_string_representation(*args, **kwargs)

    def compute_string_representation(
        self,
        coeff_fmt_str: str = "{:.2}",
        factor_fmt_str: str = "x_{dim}^{exp}",
        *args,
        **kwargs,
    ) -> str:
        representation = "[#ops={}] p(x)".format(self.num_ops)
        if self.compute_representation:
            representation += " = "
            monomials = []
            for i, exp_vect in enumerate(self.exponents):
                monomial = [coeff_fmt_str.format(self.coefficients[i, 0])]
                for dim, exp in enumerate(exp_vect):
                    # show all operations, even 1 * x_i^0
                    monomial.append(factor_fmt_str.format(**{"dim": dim + 1, "exp": exp}))

                monomials.append(" ".join(monomial))

            representation += " + ".join(monomials)

        self.representation = representation
        return self.representation

    def eval(self, x: TYPE_1D_FLOAT, rectify_input: bool = False) -> float:
        """computes the value of the polynomial at query point x

        makes use of fast ``Numba`` just in time compiled functions

        Args:
            x: ndarray of floats with shape = [self.dim] representing the query point
            rectify_input: bool, default=False
                whether to convert coefficients and exponents into compatible numpy arrays
                with this set to True, the query point x can be given in standard python arrays

        Returns:
             the value of the polynomial at point x

        Raises:
            TypeError: if x is not given as ndarray of dtype float
            ValueError: if x does not have the shape ``[self.dim]``
        """
        if rectify_input:
            x = rectify_query_point(x)
        validate_query_point(x, self.dim)

        return naive_eval(x, self.coefficients.flatten(), self.exponents)



import ctypes
import os
import shutil
import subprocess
from pathlib import Path
from typing import Tuple

from multivar_horner.classes.factorisation import BasePolynomialNode
from multivar_horner.classes.helpers import FactorContainer

if os.name == "nt":
    COMPILED_C_ENDING = ".dll"
else:
    COMPILED_C_ENDING = ".so"

DOUBLE = "double"
# array for evaluation results of both scalar and monomial factors
FACTORS = "f"
COEFFS = "c"
EVAL_FCT = "eval"
C_TYPE_DOUBLE = ctypes.c_double


def write_c_file(
    nr_coeffs: int,
    nr_dims: int,
    path_out: Path,
    factorisation: Tuple[BasePolynomialNode, FactorContainer],
    verbose: bool = False,
) -> int:
    tree, factor_container = factorisation

    # count and return the required amount of operations
    num_ops = 0

    instr = "#include <math.h>\n"
    # NOTE: the coefficient array will be used to store intermediary results
    # -> copy to use independent instance of array (NO pointer to external array!)
    func_def = f"{DOUBLE} {EVAL_FCT}({DOUBLE} x[{nr_dims}], {DOUBLE} {COEFFS}[{nr_coeffs}])"
    # declare function ("header")
    instr += f"{func_def};\n"
    # function definition
    instr += f"{func_def}"
    instr += "{\n"
    # NOTE: the order of computing the values important: scalar factors, monomial factors, monomials,...
    scalars = factor_container.scalar_factors
    monomials = factor_container.monomial_factors
    nr_factors = len(factor_container)
    # initialise arrays to capture the intermediary results of factor evaluation to 0
    if nr_factors > 0:
        instr += f"{DOUBLE} {FACTORS}[{nr_factors}]= {{0.0}};\n"
    idx: int = 0
    # set the index of each factor in order to remember it for later reference
    # ATTENTION: different from the ones used in the recipes!
    for idx, scalar in enumerate(scalars):
        scalar.value_idx = idx
        instr += scalar.get_instructions(FACTORS)
        num_ops += scalar.num_ops
    for monomial in monomials:
        idx += 1
        monomial.value_idx = idx
        monomial.factorisation_idxs = [f.value_idx for f in monomial.scalar_factors]
        instr += monomial.get_instructions(FACTORS)
        num_ops += monomial.num_ops
    # compile the instructions for evaluating the Horner factorisation tree
    instr += tree.get_instructions(COEFFS, FACTORS)
    num_ops += tree.num_ops
    # return the entry of the coefficient array corresponding to the root node of the factorisation tree
    root_idx = tree.value_idx
    instr += f"return {COEFFS}[{root_idx}];\n"
    instr += "}\n"
    if verbose:
        print(f"writing in file {path_out}")
    with open(path_out, "w") as c_file:
        c_file.write(instr)
    return num_ops


def get_compiler() -> str:
    compiler = "gcc"
    if shutil.which(compiler) is None:
        compiler = "cc"
        if shutil.which(compiler) is None:
            raise ValueError("compiler (`gcc` or `cc`) not present. install one first.")
    return compiler


def compile_c_file(compiler: str, path_in: Path, path_out: Path):
    cmd = [compiler, "-shared", "-o", str(path_out), "-fPIC", str(path_in)]
    subprocess.call(cmd)
    if not path_out.exists():
        raise ValueError(f"expected compiled file missing: {path_out}")



# -*- coding:utf-8 -*-

# NOTE: if this raises SIGSEGV, update your Numba dependency

# TODO compare difference in computed values of other methods (=numerical error)
# TODO test derivative correctness

import itertools
import unittest

import numpy as np
import pytest

from multivar_horner import HornerMultivarPolynomial, HornerMultivarPolynomialOpt, MultivarPolynomial
from multivar_horner.global_settings import FLOAT_DTYPE, UINT_DTYPE
from tests.helpers import naive_eval_reference, proto_test_case, vectorize
from tests.settings import (
    COEFF_CHANGE_DATA,
    INPUT_DATA_INVALID_TYPES_CONSTRUCTION,
    INPUT_DATA_INVALID_TYPES_QUERY,
    INPUT_DATA_INVALID_VALUES_CONSTRUCTION,
    INPUT_DATA_INVALID_VALUES_QUERY,
    MAX_INP_MAGNITUDE,
    NR_TEST_POLYNOMIALS,
    VALID_TEST_DATA,
)

DEFAULT_CONSTR_KWARGS = {
    "rectify_input": True,
    "compute_representation": True,
    "verbose": True,
    "store_c_instr": True,
    "store_numpy_recipe": True,
}


# TODO add cross class tests
class PerClassTestRegular(unittest.TestCase):
    # test cases testing a single polynomial class
    # inheritance used to change test class
    class2test = MultivarPolynomial

    def test_construction_basic(self):
        """
        test the basic construction API functionalities
        :return:
        """
        print("\nTESTING CONSTRUCTION API...")
        cls = self.class2test

        coefficients = [5.0, 1.0, 2.0, 3.0]
        exponents = [[0, 0, 0], [3, 1, 0], [2, 0, 1], [1, 1, 1]]
        coeffs_np = np.array(coefficients, dtype=FLOAT_DTYPE).reshape(4, 1)
        exp_np = np.array(exponents, dtype=UINT_DTYPE)

        p1 = cls(coeffs_np, exp_np, compute_representation=False)
        p2 = cls(coeffs_np, exp_np, compute_representation=True)
        # both must have a string representation
        # [#ops=10] p(x)
        # [#ops=10] p(x) = 5.0 x_1^0 x_2^0 x_3^0 + 1.0 x_1^3 x_2^1 x_3^0 + 2.0 x_1^2 x_2^0 x_3^1 + 3.0 x_1^1 x_2^1 x_3^1
        assert len(str(p1)) < len(str(p2))
        assert str(p1) == p1.representation

        return_str_repr = p1.compute_string_representation(coeff_fmt_str="{:1.1e}", factor_fmt_str="(x{dim} ** {exp})")
        # the representation should get updated
        assert return_str_repr == p1.representation

        # converting the input to the required numpy data structures automatically:
        p3 = cls(
            coefficients,
            exponents,
            rectify_input=True,
            compute_representation=True,
            keep_tree=True,
        )
        # the data structures should be stored as the desired numpy array
        np.array_equal(p3.coefficients, coeffs_np)
        np.array_equal(p3.exponents, exp_np)

        # partial derivative:
        # TODO own full test
        p3_deriv = p3.get_partial_derivative(2, compute_representation=True)
        # NOTE: partial derivatives themselves will be instances of the same parent class
        assert p3_deriv.__class__ is p3.__class__

        grad = p3.get_gradient(compute_representation=True)
        # TODO own full test
        # partial derivative for every dimension
        assert len(grad) == p3.dim
        print("OK.\n")

    def test_eval_api(self):
        print("\nTESTING EVALUATION API...")
        keys = ["rectify_input", "validate_input"]
        vals = [True, False]
        max_dim = 3
        degree_range = range(1, 4)
        exponents = np.array(list(itertools.product(degree_range, repeat=max_dim)), dtype=UINT_DTYPE)
        coeffs = np.random.rand(exponents.shape[0])
        X = np.random.rand(NR_TEST_POLYNOMIALS, max_dim)
        p_ref = naive_eval_reference(X, exponents, coeffs)
        for kwargs in [dict(zip(keys, tf)) for tf in itertools.product(*[vals] * len(keys))]:
            p_mv = vectorize(self.class2test(coeffs[:, None], exponents, **kwargs))
            assert np.allclose(p_ref, p_mv(X))

        print("OK.\n")

    def test_eval_compare_numpy_1d(self):
        print("TEST COMPARISON TO NUMPY 1D POLYNOMIAL EVALUATION...")
        x = MAX_INP_MAGNITUDE * ((np.random.rand(NR_TEST_POLYNOMIALS) * 2) - 1)
        for ncoeff in range(1, 8):
            exponents = np.arange(ncoeff, dtype=UINT_DTYPE).reshape(ncoeff, 1)
            coeffs = np.random.rand(ncoeff)
            # p_np = lambda x: np.polyval(coeffs[::-1], x)
            p_np = np.polynomial.Polynomial(coeffs)
            p_mv = vectorize(self.class2test(coeffs[:, None], exponents))
            results_multivar_horner = p_mv(x)
            results_numpy = p_np(x)
            assert np.allclose(results_numpy, results_multivar_horner)

        print("OK.\n")

    def test_eval_compare_vandermonde_nd(self):
        print("TEST COMPARISON TO VANDERMONDE POLYNOMIAL EVALUATION...")
        degree_range = range(1, 4)
        dim_range = range(1, 4)
        for dim, deg in itertools.product(dim_range, degree_range):
            exponents = np.array(list(itertools.product(range(deg), repeat=dim)), dtype=UINT_DTYPE)
            coeffs = np.random.rand(exponents.shape[0])
            X = np.random.rand(NR_TEST_POLYNOMIALS, dim)
            p_ref = naive_eval_reference(X, exponents, coeffs)
            p_mv = vectorize(self.class2test(coeffs[:, None], exponents))
            assert np.allclose(p_ref, p_mv(X))

        print("OK.\n")

    def test_eval_cases(self):
        print("\nTESTING STANDARD EVALUATION CASES...")
        cls = self.class2test

        def cmp_value_fct(inp):
            coeff, exp, x = inp
            x = np.array(x).T
            poly = cls(coeff, exp, **DEFAULT_CONSTR_KWARGS)
            print(cls, poly)
            res = poly(x, rectify_input=True)
            return res

        proto_test_case(VALID_TEST_DATA, cmp_value_fct)
        print("OK.\n")

    def test_invalid_input_detection(self):
        print("\n\nTEST INVALID INPUT DETECTION...")
        cls = self.class2test

        def construction_test(coeff, exp, expected_error, rectify: bool = False):
            with pytest.raises(expected_error):
                cls(coeff, exp, rectify_input=rectify)

        def construction_should_raise(data, expected_error):
            for inp, _ in data:
                coeff, exp, x = inp
                construction_test(coeff, exp, expected_error)

        construction_should_raise(INPUT_DATA_INVALID_TYPES_CONSTRUCTION, TypeError)
        construction_should_raise(INPUT_DATA_INVALID_VALUES_CONSTRUCTION, ValueError)

        # input rectification with negative exponents should raise a ValueError:
        coeff, exp = [3.2], [[0, 3, -4]]
        construction_test(coeff, exp, ValueError, rectify=True)

        def query_should_raise(data, expected_error):
            for inp, _ in data:
                coeff, exp, x = inp

                # NOTE: construction must not raise an error:
                p = cls(coeff, exp, rectify_input=False)
                with pytest.raises(expected_error):
                    p(x, rectify_input=False)

        query_should_raise(INPUT_DATA_INVALID_TYPES_QUERY, TypeError)
        query_should_raise(INPUT_DATA_INVALID_VALUES_QUERY, ValueError)
        print("OK.")

    def test_change_coefficients(self):
        print("\nTEST CHANGING COEFFICIENTS...")
        # Test if coefficients can actually be changed, representation should change accordingly
        cls = self.class2test

        def change_coeff_same_values(inp):
            # changing the coefficients while using the same coeff values should not alter the evaluation results
            # (reuse test data)
            coeff, exp, x = inp
            poly = cls(coeff, exp, rectify_input=True, compute_representation=True)
            poly = poly.change_coefficients(
                coeff,
                rectify_input=True,
                compute_representation=True,
            )
            print(poly)
            res = poly.eval(x, rectify_input=True)
            return res

        proto_test_case(VALID_TEST_DATA, change_coeff_same_values)

        def change_coeffs_fct(inp):
            print("\n")
            coeffs1, exp, x, coeffs2 = inp
            print(x)
            poly = cls(coeffs1, exp, **DEFAULT_CONSTR_KWARGS)
            print(poly)
            poly = poly.change_coefficients(
                coeffs2,
                rectify_input=True,
                compute_representation=True,
            )
            print(poly)
            res = poly(x, rectify_input=True)
            return res

        # this also tests if the factorisation tree, can still be accessed after changing the coefficients
        # representation would otherwise be empty
        # keep_tree has to be True
        proto_test_case(COEFF_CHANGE_DATA, change_coeffs_fct)
        print("OK.\n")


class PerClassTestHorner(PerClassTestRegular):
    class2test = HornerMultivarPolynomial

    def test_c_eval(self):
        print("\nTESTING C INSTRUCTION EVALUATION CASES (HORNER)...")
        cls = self.class2test

        def cmp_value_fct(inp):
            coeff, exp, x = inp
            x = np.array(x).T
            poly = cls(coeff, exp, **DEFAULT_CONSTR_KWARGS)
            print(cls, poly)
            res = poly._eval_c(x)
            return res

        print("\nTESTING EVALUATION CASES...")
        proto_test_case(VALID_TEST_DATA, cmp_value_fct)
        print("OK.\n")

    def test_recipe_eval(self):
        print("\nTESTING RECIPE EVALUATION CASES (HORNER)...")
        cls = self.class2test

        def cmp_value_fct(inp):
            coeff, exp, x = inp
            x = np.array(x).T
            poly = cls(coeff, exp, **DEFAULT_CONSTR_KWARGS)
            print(cls, poly)
            res = poly._eval_recipe(x)
            return res

        proto_test_case(VALID_TEST_DATA, cmp_value_fct)
        print("OK.\n")


class PerClassTestOptimal(PerClassTestRegular):
    class2test = HornerMultivarPolynomialOpt


if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(PerClassTestRegular)
    unittest.TextTestRunner(verbosity=2).run(suite)
    unittest.main()



# -*- coding:utf-8 -*-
from multivar_horner.classes.abstract_poly import load_pickle
from multivar_horner.classes.horner_poly import HornerMultivarPolynomial, HornerMultivarPolynomialOpt
from multivar_horner.classes.regular_poly import MultivarPolynomial

__all__ = ["HornerMultivarPolynomial", "MultivarPolynomial", "HornerMultivarPolynomialOpt", "load_pickle"]



# -*- coding:utf-8 -*-

import abc
import pickle
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List

import numpy as np

from multivar_horner.global_settings import BOOL_DTYPE, DEBUG, DEFAULT_PICKLE_FILE_NAME, TYPE_1D_FLOAT, TYPE_2D_INT
from multivar_horner.helper_fcts import (
    rectify_coefficients,
    rectify_construction_parameters,
    validate_coefficients,
    validate_construction_parameters,
)


# TODO docstring attributes in parent class, automatic inheritance in sphinx autodoc api docu
#   -> prevent duplicate docstring in MultivarPolynomial
class AbstractPolynomial(ABC):
    """an abstract class for representing a multivariate polynomial"""

    # prevent dynamic attribute assignment (-> safe memory)
    # FIXME: creates duplicate entries in Sphinx autodoc
    __slots__ = [
        "compute_representation",
        "coefficients",
        "euclidean_degree",
        "exponents",
        "num_monomials",
        "num_ops",
        "dim",
        "maximal_degree",
        "total_degree",
        "unused_variables",
        "representation",
        "verbose",
        "_hash_val",
    ]

    def __init__(
        self,
        coefficients: TYPE_1D_FLOAT,
        exponents: TYPE_2D_INT,
        rectify_input: bool = False,
        compute_representation: bool = False,
        verbose: bool = False,
    ):
        self._hash_val: int
        self.verbose: bool = verbose
        self.compute_representation: bool = compute_representation

        if rectify_input:
            coefficients, exponents = rectify_construction_parameters(coefficients, exponents)
        validate_construction_parameters(coefficients, exponents)
        self.coefficients: np.ndarray = coefficients
        self.exponents: np.ndarray = exponents

        self.num_monomials: int = self.exponents.shape[0]
        self.dim: int = self.exponents.shape[1]
        self.unused_variables = np.where(~np.any(self.exponents.astype(BOOL_DTYPE), axis=1))[0]
        self.total_degree: int = np.max(np.sum(self.exponents, axis=0))
        self.euclidean_degree: float = np.max(np.linalg.norm(self.exponents, ord=2, axis=0))
        self.maximal_degree: int = np.max(self.exponents)
        self.num_ops: int = 0
        self.representation: str

    def __str__(self):
        return self.representation

    def __repr__(self):
        return self.representation

    def __call__(self, *args, **kwargs) -> float:
        return self.eval(*args, **kwargs)

    def __hash__(self):
        """
        compare polynomials (including their factorisation) based on their properties
        NOTE: coefficients can be changed
        without affecting the fundamental properties of the polynomial (factorisation)
        NOTE: optimal factorisations might be different from the ones found with the default approach

        Returns: an integer encoding the fundamental properties of the polynomial including its factorisation
        """
        try:
            return self._hash_val
        except AttributeError:
            # lazy initialisation: compute the expensive hash value only once on demand
            props = (self.dim, self.num_monomials, *self.exponents.flatten())
            self._hash_val = hash(props)
        return self._hash_val

    def __eq__(self, other):
        """
        Returns: true when ``other`` is of the same class and has equal properties (encoded by hash)
        """
        if not isinstance(other, self.__class__):
            return False
        # we consider polynomials equal when they share their properties (-> hash)
        return hash(self) == hash(other)

    def print(self, *args):
        if self.verbose:
            print(*args)

    @abc.abstractmethod
    def compute_string_representation(self, *args, **kwargs) -> str:
        """computes a string representation of the polynomial and sets self.representation

        Returns:
            a string representing this polynomial instance
        """
        ...

    def export_pickle(self, path: str = DEFAULT_PICKLE_FILE_NAME):
        self.print(f'storing polynomial in file "{path}"')
        with open(path, "wb") as f:
            pickle.dump(self, f)

    # TODO test
    def get_partial_derivative(self, i: int, *args, **kwargs) -> "AbstractPolynomial":
        """retrieves a partial derivative

        Note:
            all given additional arguments will be passed to the constructor of the derivative polynomial

        Args:
            i: dimension to derive with respect to.
                ATTENTION: dimension counting starts with 1 (i >= 1)

        Returns:
            the partial derivative of this polynomial wrt. the i-th dimension
        """

        assert 0 < i <= self.dim, "invalid dimension i given"
        coord_index = i - 1
        # IMPORTANT: do not modify the stored coefficient and exponent arrays of self!
        # set all the coefficients not depending on the i-th coordinate to 0
        # this simply means not adding them to the list of coefficients of the new polynomial class instance
        active_rows = np.where(self.exponents[:, coord_index] >= 1)[0]

        new_coefficients = self.coefficients[active_rows]
        new_exponents = self.exponents[active_rows, :]

        if DEBUG:  # TODO move to tests
            assert new_coefficients.shape[0] == new_exponents.shape[0]
            assert new_coefficients.shape[1] == 1 and len(new_coefficients.shape) == 2
            assert new_exponents.shape[1] == self.dim and len(new_exponents.shape) == 2

        # multiply the coefficients with the exponent of the i-th coordinate
        # f(x) = a x^b
        # f'(x) = ab x^(b-1)
        new_coefficients = np.multiply(new_coefficients.flatten(), new_exponents[:, coord_index])
        new_coefficients = new_coefficients.reshape(-1, 1)

        # reduce the the exponent of the i-th coordinate by 1
        new_exponents[:, coord_index] -= 1

        # must call the proper constructor method also for inherited classes
        return self.__class__(new_coefficients, new_exponents, *args, **kwargs)

    def get_gradient(self, *args, **kwargs) -> List["AbstractPolynomial"]:
        """
        Note:
            all arguments will be passed to the constructor of the derivative polynomials

        Returns:
             the list of all partial derivatives
        """
        return [self.get_partial_derivative(i, *args, **kwargs) for i in range(1, self.dim + 1)]

    def change_coefficients(
        self,
        coefficients: TYPE_1D_FLOAT,
        rectify_input: bool = False,
        compute_representation: bool = False,
        in_place: bool = False,
        *args,
        **kwargs,
    ) -> "AbstractPolynomial":

        if rectify_input:
            coefficients = rectify_coefficients(coefficients)

        validate_coefficients(coefficients)

        assert len(coefficients) == self.num_monomials

        if in_place:
            updated_poly = self
        else:
            updated_poly = deepcopy(self)

        updated_poly.compute_representation = compute_representation

        updated_poly.coefficients = coefficients
        updated_poly.compute_string_representation(*args, **kwargs)
        return updated_poly

    @abstractmethod
    def eval(
        self,
        x: TYPE_1D_FLOAT,
        rectify_input: bool = False,
    ) -> float:
        pass


def load_pickle(path: str = DEFAULT_PICKLE_FILE_NAME) -> AbstractPolynomial:
    print('importing polynomial from file "{}"'.format(path))
    with open(path, "rb") as f:
        return pickle.load(f)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['multivar_horner', 'multivar_horner.classes']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22,<2.0']

extras_require = \
{'numba': ['numba>=0.55.2,<0.56.0']}

setup_kwargs = {
    'name': 'multivar-horner',
    'version': '3.0.4',
    'description': 'python package implementing a multivariate Horner scheme for efficiently evaluating multivariate polynomials',
    'long_description': '===============\nmultivar_horner\n===============\n\n\n.. image:: https://travis-ci.org/jannikmi/multivar_horner.svg?branch=master\n    :alt: CI status\n    :target: https://travis-ci.org/jannikmi/multivar_horner\n\n.. image:: https://readthedocs.org/projects/multivar_horner/badge/?version=latest\n    :alt: documentation status\n    :target: https://multivar_horner.readthedocs.io/en/latest/?badge=latest\n\n.. image:: https://img.shields.io/pypi/wheel/multivar-horner.svg\n    :target: https://pypi.python.org/pypi/multivar-horner\n\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n\n.. image:: https://pepy.tech/badge/multivar-horner\n    :alt: Total PyPI downloads\n    :target: https://pepy.tech/project/multivar-horner\n\n.. image:: https://img.shields.io/pypi/v/multivar_horner.svg\n    :alt: latest version on PyPI\n    :target: https://pypi.python.org/pypi/multivar-horner\n\n.. image:: https://joss.theoj.org/papers/0b514c6894780f3cc81ed88c141631d4/status.svg\n    :alt: JOSS status\n    :target: https://joss.theoj.org/papers/0b514c6894780f3cc81ed88c141631d4\n\n.. image:: https://zenodo.org/badge/155578190.svg\n   :target: https://zenodo.org/badge/latestdoi/155578190\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n\n\n``multivar_horner`` is a python package implementing a multivariate\n`Horner scheme ("Horner\'s method", "Horner\'s rule") <https://en.wikipedia.org/wiki/Horner%27s_method>`__\nfor efficiently evaluating multivariate polynomials.\n\n\nQuick Guide:\n\n::\n\n\n    pip install multivar_horner\n\n\nFor efficiency this package is compiling the instructions required for polynomial evaluation to C by default.\nIf you don\'t have a C compiler (``gcc`` or ``cc``) installed you also need to install numba for using an alternative method:\n\n::\n\n\n    pip install multivar_horner[numba]\n\n\nSimple example:\n\n.. code-block:: python\n\n    import numpy as np\n    from multivar_horner import HornerMultivarPolynomial\n\n    # input parameters defining the polynomial\n    #   p(x) = 5.0 + 1.0 x_1^3 x_2^1 + 2.0 x_1^2 x_3^1 + 3.0 x_1^1 x_2^1 x_3^1\n    coefficients = np.array([[5.0], [1.0], [2.0], [3.0]], dtype=np.float64)\n    exponents = np.array([[0, 0, 0], [3, 1, 0], [2, 0, 1], [1, 1, 1]], dtype=np.uint32)\n\n    # [#ops=7] p(x) = x_1 (x_1 (x_1 (1.0 x_2) + 2.0 x_3) + 3.0 x_2 x_3) + 5.0\n    horner_polynomial = HornerMultivarPolynomial(coefficients, exponents)\n    x = np.array([-2.0, 3.0, 1.0], dtype=np.float64)\n    p_x = horner_polynomial(x)\n\n\nFor more refer to the `documentation <https://multivar_horner.readthedocs.io/en/latest/>`__.\n\n\nAlso see:\n`GitHub <https://github.com/jannikmi/multivar_horner>`__,\n`PyPI <https://pypi.python.org/pypi/multivar_horner/>`__,\n`arXiv paper <https://arxiv.org/abs/2007.13152>`__\n',
    'author': 'jannikmi',
    'author_email': 'github@michelfe.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://multivar-horner.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
